
/*
   helgraf2.c
   
   Graphics routines for helicon.
   
   Author: Mark E. Nash
   Date:   July, 1988
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<graph.h>
#include	<malloc.h>
#include	<math.h>
#include	<string.h>
#include	<process.h>
#include	<conio.h>
#include	<io.h>
#include	<dos.h>
#include	<bios.h>
#include	<macs.h>
#include	"helimacs.h"
#include	"helimess.h"
#include	"helstruc.h"
#include	"helfdef.h"
#include	"egafon.h"

/* externals */
extern char msg[];				/* in heliglob.h */
extern struct grafwindow *top_window;		/* in heliglob.h */
extern unsigned delcon;				/* in heliglob.h */
extern short cxpos, cypos;			/* in heliglob.h */
extern unsigned mouse;				/* in heliglob.h */
extern unsigned m1, m2, m3, m4;			/* in heliglob.h */
	
	
	
int gtwcurpos(wp, col, row)
	struct grafwindow *wp;
	unsigned short col, row;
/* 
   gtwcurpos() positions the text cursor in a graphics mode text window.
   
   Entry: wp is the graphics mode text window to use.
          col = the column position to move the cursor to.
	  row = the row position to move the cursor to.
	  col >= 0, row >= 0.
	  
   Exit:  If no errors, the text cursor has been moved to (row, col).
          Returns NOERR.
*/
	{
	if (wp != NULL)
		{
		if (col >= 0 && col <= wp->maxcols)
			wp->curcol = col;
		if (row >= 0 && row <= wp->maxrows)
			wp->currow = row;
		}
		
	return(NOERR);
	}

	
	
int gtwgetpos(wp, col, row)
	struct grafwindow *wp;
	unsigned short *col, *row;
/* 
   gtwgetpos() gets the position of the the text cursor in a graphics mode 
   text window.
   
   Entry: wp is the graphics mode text window to use.
          col = the column position that will be returned.
	  row = the row position that will be returned.
	  
   Exit:  If no errors, the text cursor has been moved to (row, col).
          Returns NOERR.
*/
	{
	if (wp != NULL)
		{
		*col = wp->curcol;
		*row = wp->currow;
		}
	return(NOERR);
	}
 


int eputs_window(wp, str, color, center)
	struct grafwindow *wp;
	char str[];
	short color;
	short center;
/* 
   eputs_window() puts a string in the specified graphics mode text window at
   the current cursor position.
   
   Entry: wp is the window to write to.
          str is the string to write.
	  color = color of the string.
	  center = TRUE if we are to center each row, otherwise FALSE.
	  
   Exit:  The string specified has been written to the specified window.
          Returns NOERR.
*/
	{
	unsigned ctr, sctr, sctr1;	/* count in the string */
	char	wstr[WRLEN + 1];	/* string to write to a window */
	char	*beg;			/* pointers for parsing */
	double	x, y;			/* scaled position of character */


	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "eputs_window()");
		err_window(msg);
		return(ERR);
		}
		
	if (wp->wintype == GRAFWIN)
		return(NOERR);
		
/* turn off mouse cursor */
	MOUSE(2, 0, 0, 0);
		
	sctr = 0;
	while (str[sctr] != '\0')
		{
/* start a buffer */
		wstr[0] = '\0';

/* skip over leading spaces on all but first line */
		if (sctr != 0)
			while (str[sctr] == ' ')
				sctr++;
				
		beg = &(str[sctr]);
		
/* get a buffer's worth of characters */
		while (str[sctr] != '\0' && (&(str[sctr]) - beg) <= 
								wp->maxcols)
			sctr++;
			
/* break on a space if possible */
		sctr1 = sctr;
		while (str[sctr] != '\0' && str[sctr] != ' ' && 
						(&(str[sctr]) - beg) > 0)
			sctr--;
		if ((&(str[sctr]) - beg) == 0)
			sctr = sctr1;
			
/* center if necessary */
		if (center == TRUE)
			{
			for (ctr=0; ctr<((wp->maxcols+1) - 
						(&(str[sctr])-beg))/2; ctr++)
				strcat(wstr, " ");
			}
			
/* write the string */
		strncat(wstr, beg, (unsigned)(&(str[sctr]) - beg));

/* set up the ega registers */
		egafunc(color, XOR);		/* XOR is a dummy function */
/* display the string */		
		ctr = 0;
		y = wp->npy * (double)VPIXCHAR * 
					(double)(wp->maxrows - wp->currow);
		while (wstr[ctr] != '\0')
			{
/* bell */
			if (wstr[ctr] == '\7')
				printf("\7");
/* new line */
			else
				if (wstr[ctr] == '\n')
					{
					wp->curcol = 0;
					wp->currow += 1;
					y = wp->npy * (double)VPIXCHAR * 
					    (double)(wp->maxrows - wp->currow);
					}
/* character */
				else
					{
					x = wp->npx * (double)HPIXCHAR * 
							(double)wp->curcol;
					win_char(wp, wstr[ctr], x, y, color);
					wp->curcol += 1;
					}
			ctr++;
			}
		wp->currow += 1;
		wp->curcol = 0;
		}
/* reset the ega registers */
	egareset();

/* turn mouse back on */
	MOUSE(1, 0, 0, 0);
		
	return(NOERR);
	}



int eputs_ega(str, col, row, linlen, color, center)
	char str[];
	int row, col;
	unsigned linlen;
	short color, center;
/* 
   eputs_ega() writes a string to the ega graphics display.
   
   Entry: str is the string to write.
          (col, row) is where to start writing the string (logical coords).
	  linlen is the line length (for centering).
	  color = color of the string.
	  center = TRUE if we are to center each row, otherwise FALSE.
	  
   Exit:  The string specified has been written to the ega graphics display.
          Returns NOERR.
*/
	{
	unsigned ctr, sctr, sctr1;	/* count in the string */
	unsigned curcol, currow;	/* current row and column */
	char	wstr[WRLEN + 1];	/* string to write to a window */
	char	*beg;			/* pointers for parsing */


/* turn off mouse cursor */
	MOUSE(2, 0, 0, 0);
		
	sctr = 0;
	while (str[sctr] != '\0')
		{
/* start a buffer */
		wstr[0] = '\0';

/* skip over leading spaces on all but first line */
		if (sctr != 0)
			while (str[sctr] == ' ')
				sctr++;
				
		beg = &(str[sctr]);
		
/* get a buffer's worth of characters */
		while (str[sctr] != '\0' && (&(str[sctr]) - beg) <= linlen)
			sctr++;
			
/* break on a space if possible */
		sctr1 = sctr;
		while (str[sctr] != '\0' && str[sctr] != ' ' && 
						(&(str[sctr]) - beg) > 0)
			sctr--;
		if ((&(str[sctr]) - beg) == 0)
			sctr = sctr1;
			
/* center if necessary */
		if (center == TRUE)
			{
			for (ctr=0; ctr<(linlen - (&(str[sctr])-beg))/2; ctr++)
				strcat(wstr, " ");
			}
			
/* write the string */
		strncat(wstr, beg, (unsigned)(&(str[sctr]) - beg));
		
/* set up the ega registers */
		egafunc(color, XOR);		/* XOR is a dummy function */
/* display the string */
		ctr = 0;
		currow = row;
		curcol = col;
		while (wstr[ctr] != '\0')
			{
/* bell */
			if (wstr[ctr] == '\7')
				printf("\7");
/* new line */
			else
				if (wstr[ctr] == '\n')
					{
					curcol = col;
					currow += VPIXCHAR;
					}
/* character */
				else
					{
					ega_char(wstr[ctr], curcol, currow);
					curcol += HPIXCHAR;
					}
			ctr++;
			}
		currow += VPIXCHAR;
		curcol = col;
		}
		
/* reset the ega */
	egareset();

/* turn mouse back on */
	MOUSE(1, 0, 0, 0);
		
	return(NOERR);
	}
	
	
	
int win_char(wp, c, x, y, color)
	struct grafwindow *wp;
	unsigned char c;
	double x, y;
	short color;
/* 
   win_char() puts the character at the specified window on an EGA display.
   
   Entry: wp is the window to write to.
          c = the character code (0-127).
          x = x coord. at the middle of the character cell.
          y = y coord. at the middle of the character cell.
	  color = color of the character.
	  
   Exit:  Returns NOERR if no errors, else ERR.
*/
	{
	int	row, col;			/* logical coords */


	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "win_char()");
		err_window(msg);
		return(ERR);
		}
		
/* scale the point */
	y = wp->ysmax + wp->ysmin - y;
	if (wp->npx != 0.0 && wp->npy != 0.0)
		{
		col = (short)(((x - wp->xsmin) / wp->npx) + .5) + wp->wbox.x1;
		row = (short)(((y - wp->ysmin) / wp->npy) + .5) + wp->wbox.y1;
	
/* get out if character cannot be drawn at this position. */
		if (row < (wp->wbox.y1 + 4) || row > (wp->wbox.y2 - 4) ||
					col < (wp->wbox.x1 + 4) || col > 
							(wp->wbox.x2 - 4))
			return(NOERR);

		if (ega_char(c, col, row) == ERR)
			return(ERR);
		}

	return(NOERR);
	}
	


int ega_char(c, col, row)
	unsigned char c;
	int col, row;
/*
   ega_char() puts a character on the ega graphics mode screen.
   
   Entry: The color and write enable have been set for the ega.
          c = the character code (0-127).
          row is the logical coord. of the center of the char row position.
          col is the logical coord. of the center of the char col position.
	  
   Exit:  Returns NOERR if no errors, else ERR.
   
   Note:  Should call egareset() when done with this routine.
          This routine is optmized for the ega byte-wide screen writing.
*/
	{
	int 	pr;				/* row pixel position */
	register int mask1, mask2;		/* for setting both sections */
	unsigned char exist_color;		/* for dummy read */
	unsigned long base, basemax, basinc;	/* point to ega memory */
	char	begchk, endchk;			/* mask shift */
	struct xycoord xy;			/* for coord shift */
	int	onebyte;			/* cross byte boundary? */
	static char fake = TRUE;		/* to fake-out the compiler */


	xy = _getphyscoord(col, row);
	col = xy.xcoord;
	row = xy.ycoord;
	
/* initialize things */
	col = col - HPIXCHAR / 2;	/* because coords are centered */
	begchk = (char)(col % 8);
	endchk = (char)(8 - ((col + HPIXCHAR) % 8));
	
	if (begchk == 0)
		onebyte = TRUE;
	else
		onebyte = FALSE;
						
	pr = row - VPIXCHAR / 2 + 1;
	base = EGABASE + (long)pr * WIDTH + (long)(col / 8);
	basemax = base + VPIXCHAR * EGABR;
	basinc = EGABR - 1;
	
	_disable();			/* disable interrupts here */
	
	for (; base<basemax; base+=basinc, pr++)
		{
/* set the fraction of the first byte */
		mask1 = ega_font[(int)c][pr - row + VPIXCHAR / 2 - 1];
		mask2 = mask1;
		mask1 >>= begchk;
		exist_color = *(char far *)base;/* set up ega registers */
		OUTINDEX(8, mask1);		/* set the mask to use */
		OUTINDEX(3, OR);		/* or */
		*(char far *)base &= fake;	/* do the write */
		OUTINDEX(3, AND);		/* and */
		*(char far *)base &= fake;	/* do the write */

/* set the fraction of the second byte */
		base++;
		if (!onebyte)
			{
			mask2 <<= endchk;
			exist_color = *(char far *)base;/* set up ega regs */
			OUTINDEX(8, mask2);	/* set the mask to use */
			OUTINDEX(3, OR);		/* or */
			*(char far *)base &= fake;	/* do the write */
			OUTINDEX(3, AND);		/* and */
			*(char far *)base &= fake;	/* do the write */
			}
		}
	
	_enable();		/* re-enable interrupts */
	return(NOERR);
	}
	
	

void egafunc(color, func)
	int color, func;
/*
   egafunc() sets up the ega registers with the given color and function.
   
   Entry: color is the color mask to use.
	  func is the function XOR, OR, AND.
	  
   Exit:  The color and function registers of the ega have been set.
*/
	{
	_disable();			/* disable interrupts here */
	OUTINDEX(0, color);		/* set the color */
	OUTINDEX(1, ENABLE);		/* enable a write */
	OUTINDEX(3, func);		/* tell the function to use */
	_enable();			/* re-enable interrupts */
	return;
	}
	
	
	
void egaline(col1, col2, row)
	int col1, col2, row;
/*
   egapixel() sets an ega row to the value set by egafunc().
   
   Entry: col1 is the beginning phys coord of the row.
          col2 is the ending phys coord of the row.
	  row is the phys coord. of the row.
          	  
   Exit:  If the row is within the display bounds, then it has been set
		according to color and func set in egafunc().  If not, 
		you have problems, since there is no error checking.

   Note:  This function is optimized to take advantage of the byte-wide write
          to the ega.
*/
	{
	register int mask;			/* for setting pixel */
	unsigned char exist_color;		/* for dummy read */
	char far *base1, *base2;		/* point to ega memory */
	int 	begchk, endchk;			/* for checking ends of row */
	int	onebyte;			/* test for one-byte row */
	static char fake = TRUE;		/* to fake-out the compiler */
	
	
/* initialize things */
	begchk = (char)(col1 % 8);
	endchk = (char)(col2 % 8);

	base1 = (char far *)(EGABASE + ((long)row * WIDTH + 
						((long)col1 / 8L)));
	base2 = (char far *)(EGABASE + ((long)row * WIDTH + 
						((long)col2 / 8L)));
						
/* check to see if this is a one-byte line */
	if (base1 == base2)
		onebyte = TRUE;
	else
		onebyte = FALSE;


/* set the fraction of the first byte */
	mask = 0xFF;
	mask >>= begchk;
	if (onebyte)
		mask <<= endchk;

	_disable();			/* disable ints */
	
	exist_color = *base1;		/* set up ega registers */
	OUTINDEX(8, mask);		/* set the mask to use */
	*base1 &= fake;			/* do the write */

/* do the next bytes */
	mask = 0xFF;
	OUTINDEX(8, mask);			/* set the mask to use */
	for (base1++; base1<base2; base1++)
		{
		exist_color = *base1;		/* set up ega registers */
		*base1 &= fake;			/* do the write */
		}
		
/* set the fraction of the last byte */
	if (!onebyte)
		{
		mask >>= endchk + 1;
		mask = ~mask;
		exist_color = *base1;		/* set up ega registers */
		OUTINDEX(8, mask);		/* set the mask to use */
		*base1 &= fake;			/* do the write */
		}

	_enable();			/* enable interrupts */
	return;	
	}
	
	
	
void egareset()
/*
   egareset() resets the ega registers, which must be done before exiting
   to dos.
*/
	{
	_disable();		/* disable interrupts */
	OUTINDEX(0, 0);			/* reset register */
	OUTINDEX(1, 0);			/* reset enable */
	OUTINDEX(3, 0);			/* reset logic and bit rotation */
	OUTINDEX(8, TRUE);		/* reset bit mask */
	_enable();		/* re-enable interrupts */
	return;
	}
	
	
	
int win_cur(wp)
	struct grafwindow *wp;
/*
   win_cur() xor's a cursor at the given position.
   
   Entry: The position was set by gtwcurpos().
   
   Exit:  A cursor has been xor'ed at the current position.
*/
	{
	int	scol, ecol, crow;	/* cursor position phys coords */
	struct xycoord xy;		/* coord conversion */
	
	
	xy = _getphyscoord(wp->wbox.x1, wp->wbox.y1);
	
/* calculate the cursor's physical coords */
	scol = xy.xcoord + wp->curcol * HPIXCHAR + HPIXCHAR / 2;
	ecol = scol + HPIXCHAR;
	crow = xy.ycoord + (wp->currow + 1) * VPIXCHAR  + VPIXCHAR / 2 + 1;

	egafunc(BRIGHTWHITE, XOR);
	egaline(scol, ecol, crow);
	egareset();
	return(NOERR);
	}
	
	
	
int plotdat(class)
	int class;
/*
   plotdat() makes frequency data out of the raw data and plots it.
   
   Entry: class is the distribution class = DISCRETE, CONTINUOUS or CHOOSE.
	  The raw data list is complete.

   Exit:  The data is in frequency form and histogram plots have been made.
*/
	{
	struct bin *data;			/* bin data */
	static double sxmin, symin,
		      sxmax, symax;		/* data window scale */
	int	numbins;			/* number of bins */
	int	lrow, lcol;			/* for labels */
	extern struct rawcell *rd_top;		/* in heliglob.h */
	extern struct topbin *bintop;		/* in heliglob.h */
	extern char tboxes[NUMTBXS][TBOXSIZE+1];/* in heliglob.h */
	extern struct menutop *menus;		/* in heliglob.h */
	extern int autoscale;			/* in heliglob.h */
	extern struct grafwindow *dw;		/* in heliglob.h */
	extern struct grafwindow *pw;		/* in heliglob.h */
	

/* check to see if raw data is there */
	if (rd_top == NULL)
		{
		err_window(D_NORAW);
		return(ERR);
		}

/* check the class */
	if (class == CHOOSE)
		if (bintop != NULL)
			class = bintop->dclass;
		else
			class = CONTINUOUS;

/* generate bins */
	free_bins(&bintop);
	numbins = atoi(tboxes[NBINS]);
	if (makebins(rd_top, &bintop, &numbins, class) == ERR)
		return(ERR);

	tboxes[NBINS][0] = '\0';
	sprintf(tboxes[NBINS], "%d", numbins);
				
	if (autoscale)
		{
		sprintf(msg, "%-g", bintop->xmin - BSCONST * (bintop->xmax - 
								bintop->xmin));
		strncpy(tboxes[TMINX], msg, TBOXSIZE);
		sprintf(msg, "%-g", -SCONST * bintop->ymax);
		strncpy(tboxes[TMINY], msg, TBOXSIZE);
		sprintf(msg, "%-g", bintop->xmax + SCONST * (bintop->xmax - 
								bintop->xmin));
		strncpy(tboxes[TMAXX], msg, TBOXSIZE);
		sprintf(msg, "%-g", bintop->ymax + SCONST * bintop->ymax);
		strncpy(tboxes[TMAXY], msg, TBOXSIZE);
		}

	sxmin = atof(tboxes[TMINX]);
	symin = atof(tboxes[TMINY]);
	sxmax = atof(tboxes[TMAXX]);
	symax = atof(tboxes[TMAXY]);

	if (scale_grafwin(dw, sxmin, symin, sxmax, symax) == ERR)
		{
		err_window(W_BADSCALE);
		return(ERR);
		}

/* draw axes and scales */
	win_axes(dw, AXISCOLOR);
	win_dscale(dw, SCALECOLOR, XAXIS);
	win_dscale(dw, SCALECOLOR, YAXIS);

/* plot histograms */
	data = bintop->start;
	while (data != NULL)
		{
		if (data->y > 0.0)
			{
			win_rect(dw, data->xval, data->y, data->xval + 
					bintop->binsize, 0.0, SCACOLOR, FALSE);
			win_point(dw, data->x, data->y, DPCOLOR);
			}
		data = data->next;
		}

/* see if drawing possible */
	if (dw->npx == 0.0 || dw->npy == 0.0)
		return(ERR);
		
/* show the mean and var */
	win_line(dw, bintop->mean, dw->ysmin, bintop->mean, dw->ysmin + 
			(dw->ysmax - dw->ysmin) * (MAJTIC+0.01), MEANCOLOR);
	
	sprintf(msg, "%s", "mean");
	lcol = dw->wbox.x1 + (int)((bintop->mean - dw->xsmin) / dw->npx) - 
						(HPIXCHAR * (strlen(msg) / 2));

/* if label goes over either end, push it back on */
	if (dw->wbox.x2 < lcol + (int)(HPIXCHAR * strlen(msg)) + HPIXCHAR/2)
		lcol = dw->wbox.x2 - (int)(HPIXCHAR*strlen(msg)-HPIXCHAR/2);
	if (dw->wbox.x1 > lcol - HPIXCHAR/2)
		lcol = dw->wbox.x1 + HPIXCHAR/2;
			
	lrow = dw->wbox.y1 + VPIXCHAR + (int)((dw->ysmax - dw->ysmin) * MAJTIC
								 / dw->npy);
	lrow = dw->wbox.y2 + dw->wbox.y1 - lrow;
	eputs_ega(msg, lcol, lrow, MESSLEN, MEANCOLOR, FALSE);
	
/* show parameters as numbers */
	gtwcurpos(pw, 0, PWROWS-2);
	sprintf(msg, "mean = %-g", bintop->mean);
	eputs_window(pw, msg, PWDCOLOR, FALSE);
	gtwcurpos(pw, 0, PWROWS-1);
	sprintf(msg, "variance = %-g", bintop->var);
	eputs_window(pw, msg, PWDCOLOR, FALSE);

	return(NOERR);
	}
	
	
	
int dislabs()
/*
   dislabs() displays the graph labels.
   
   Entry: A request to display the labels has been made.
   
   Exit:  The labels have been displayed.
*/
	{
	extern struct grafwindow *tw;		/* in heliglob.h */
	extern struct grafwindow *xw;		/* in heliglob.h */
	extern struct grafwindow *yw;		/* in heliglob.h */
	extern char tboxes[NUMTBXS][TBOXSIZE+1];/* in heliglob.h */
	int	yrow;		/* for finding start of y-axis label */
	int	ctr;				/* gp counter */
	char	onch[2];	/* for printing a character as a string */


/* output label strings */
	win_clear(tw, IBKCOLOR);
	gtwcurpos(tw, 0, 0);
	eputs_window(tw, tboxes[GTITL], GLCOLOR, TRUE);
	eputs_window(tw, tboxes[GSTITL], GLCOLOR, TRUE);
	win_clear(xw, IBKCOLOR);
	gtwcurpos(xw, 0, 0);
	eputs_window(xw, tboxes[GXAXIS], GLCOLOR, TRUE);
	yrow = (YAROWS - strlen(tboxes[GYAXIS])) / 2;
	win_clear(yw, IBKCOLOR);
	for (ctr=0; ctr<strlen(tboxes[GYAXIS]); ctr++)
		{
		gtwcurpos(yw, 0, ctr+yrow);
		onch[0] = tboxes[GYAXIS][ctr];
		onch[1] = '\0';
		eputs_window(yw, onch, GLCOLOR, TRUE);
		}
	return(NOERR);
	}
	
	
	
int geti(x1, y1, x2, y2, memptr, is)
	short x1, y1, x2, y2;
	char far **memptr;
	long *is;
/*
   geti() gets and stores an image on the ega graphics screen.
   
   Entry: (x1,y1), (x2,y2) are the image bounds.

   Exit:  The image has been stored in memory if no errors.
          memptr points to the area in memory that the image is stored in.
	  is points to the size of the memory area.
   
   Note:  The image is stored as follows:  Each byte contains a pixel 
          of information which is stored column-wise.
*/
	{
	struct xycoord xy1, xy2;		/* coord conv */
	short	px1, py1, px2, py2;		/* physical coords */
	char far *imem;				/* image pointer */
	int	row, col;			/* current pixel read */
	char	byte;				/* for building a mem byte */
	long	size;				/* size of the image */
	
	
/* turn off mouse cursor */
	MOUSE(2, 0, 0, 0);
		
/* get the size */
	size = isize(x1, y1, x2, y2);
	*is = size;
	if ((*memptr = (char far *)halloc(size, sizeof(char))) == NULL)
		{
		printf("\7");
/* turn mouse back on */
		return(NULL);
		}
/* update memory status */
	memupdt((unsigned long)size, ALLOC);
		
	imem = *memptr;
	
/* convert the coords */
	xy1 = _getphyscoord(x1, y1);
	xy2 = _getphyscoord(x2, y2);
	px1 = xy1.xcoord;
	py1 = xy1.ycoord;
	px2 = xy2.xcoord;
	py2 = xy2.ycoord;
	
/* get and store the image */
	for (col=px2; col>=px1; col--)		/* reverse column order */
		{
		for (row=py1; row<=py2; row++)
			{
			*imem = (char)getpt(col, row);
			imem++;
			}
		}

/* turn mouse back on */
	MOUSE(1, 0, 0, 0);

	return(NOERR);
	}
	
	
	
long isize(x1, y1, x2, y2)
	short x1, y1, x2, y2;
/*
   isize() gets the size of an image in bytes on the ega graphics screen.
   
   Entry: (x1,y1) and (x2,y2) are the bounds of the image in logical coords.
   
   Exit:  The image size is returned, 0 for error.
*/
	{
	long	size;		/* image size */
	
	
	size = (long)abs(x2-x1+1) * (long)abs(y2-y1+1);
	return(size);
	}
	
	
	
int ifree(memptr, size)
	char far **memptr;
	long size;
/*
   ifree() liberates the memory used by the geti() routine.
   
   Entry: *memptr points to the memory to free.
          size is the size of the memory block in bytes.
	  
   Exit:  The memory has been liberated.
*/
	{
/* free the memory */
	hfree((void huge *)(*memptr));
/* update memory status */
	memupdt((unsigned long)size, FRE);
/* check the heap */
	switch (_heapchk())
		{
		case _HEAPOK:
		case _HEAPEMPTY:
			break;
		case _HEAPBADBEGIN:
		case _HEAPBADNODE:
			printf("\7");
		}
		
	*memptr = NULL;
	return(NOERR);
	}
	
	
	
int compwin()
/*
   compwin() opens up the error comparison window.
   
   Entry: The compare menu selection has been made.
          The params[][] global is current.
   
   Exit:  The user has clicked in the compare window.
*/
	{
	struct grafwindow *cw;		/* compare window */
	struct xycoord xy;		/* for coord conv. */
	int	done;			/* check for exit */
	int	ctr;			/* gp counter */
	unsigned color;			/* text color */
	int	tag;			/* dist. with minimum error */
	double	minerr;			/* minimum error */
	extern char *fitlist[FITSIZE];	/* in heliglob.h */
	extern double params[MAXPARS+1][NUMDIST+1];	/* heliglobs.h */


/* open as a text window */
	xy = _getlogcoord(CWX1, CWY1);
	if ((cw = open_grafwin(TEXTWIN, xy.xcoord, xy.ycoord, 0, 0, 
			CWBRCOLOR, CWBKCOLOR, CWCOLS, CWROWS, TRUE)) == NULL)
		return(ERR);
		
/* find dist. with minimum error */
	tag = -1;
	minerr = -1.0;
	ctr = NULLDIST;
	while (ctr < NUMDIST && minerr == -1.0)
		{
		ctr++;
		if (params[MSE][ctr] != -1.0)
			{
			minerr = params[MSE][ctr];
			tag = ctr;
			}
		}
	if (tag != -1)
		for (ctr=tag; ctr<=NUMDIST; ctr++)
			if (params[MSE][ctr] != -1.0)
				if (params[MSE][ctr] < minerr)
					{
					minerr = params[MSE][ctr];
					tag = ctr;
					}

/* write out window data */
	gtwcurpos(cw, 0, 0);
	eputs_window(cw, "Mean Squared Error for Distributions", CWTCOLOR, 
									TRUE);
	gtwcurpos(cw, 0, 3);
	for (ctr=NULLDIST+1; ctr<=NUMDIST; ctr++)
		{
		if (params[MSE][ctr] != -1.0)
			sprintf(msg, "%-20s%-g", fitlist[ctr], 
							params[MSE][ctr]);
		else
			sprintf(msg, "%-20s", fitlist[ctr]);
			
		if (ctr == tag)
			color = CWHCOLOR;
		else
			color = CWDCOLOR;
			
		eputs_window(cw, msg, color, FALSE);
		}
	
/* wait for finish cue */
	if (mouse)
		{
/* check the mouse status */
		done = FALSE;
		while (!done)
			{
			MOUSE(3, 0, 0, 0);
			xy = _getlogcoord(m3, m4);
			cxpos = xy.xcoord;
			cypos = xy.ycoord;
			if (m2 != 0)
				if (curnrgn(&(cw->wbox)))
					done = TRUE;
			}
		}
/* wait for a return character */
	else
		while (getch() != RETURN);
	
	close_grafwin(&cw);
	return(NOERR);
	}
